using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Xml;
using System.Xml.Serialization;
using Microsoft.TeamFoundation.Build.Common;
using Microsoft.TeamFoundation.Build.Proxy;
using Microsoft.TeamFoundation.Client;
using Microsoft.TeamFoundation.VersionControl.Client;
using Microsoft.TeamFoundation.VersionControl.Common;
using Pegasus.Log4Net;
using Pegasus.Threading;
using BuildData=Microsoft.TeamFoundation.Build.Proxy.BuildData;
using BuildParameters=Microsoft.TeamFoundation.Build.Proxy.BuildParameters;
using System.Threading;

namespace BuildLibrary
{
    internal class ScanProjects : IScheduleTask
    {
        private static readonly ILog s_logger = LogManager.GetLogger( typeof( ScanProjects ) );

        #region Properties
        private readonly int m_scanInterval;
        private readonly TeamFoundationServer m_teamServer;
        private readonly VersionControlServer m_versionControl;
        private DateTime m_nextRun = DateTime.Now;
        private DateTime m_scanStartTime = DateTime.Now;
        private BuildStore m_buildStore;
        private BuildController m_controller;
        private TeamProject m_project;
        private Item m_buildFileItem;
        private Project m_projectData;
        private string m_buildType;

        private BuildStore Store
        {
            get
            {
                if( m_buildStore == null )
                {
                    m_buildStore = (BuildStore) m_teamServer.GetService( typeof( BuildStore ) );
                }
                return m_buildStore;
            }
        }

        private BuildController Controller
        {
            get
            {
                if( m_controller == null )
                {
                    m_controller = (BuildController) m_teamServer.GetService( typeof( BuildController ) );

                }
                return m_controller;
            }
        }
        #endregion

        public DateTime GetNextReoccurrence()
        {
            return m_nextRun;
        }

        public string Name
        {
            get { return "Scan Team Foundation Projects"; }
        }

        public bool ReoccurringTask
        {
            get { return true; }
        }

        public void RunTask()
        {
            m_scanStartTime = DateTime.Now;
            s_logger.InfoFormat( "Starting Project Scan at {0}", m_scanStartTime );

            foreach( TeamProject project in m_versionControl.GetAllTeamProjects( true ) )
            {
                m_project = project;
                s_logger.InfoFormat( "Processing Project {0}", project.Name );
                ScanBuildFiles();
            }

            m_nextRun = DateTime.Now.AddMinutes( m_scanInterval );
            s_logger.InfoFormat( "Completed Project Scan, next scan scheduled for {0}", m_nextRun );
        }

        public ScanProjects( TeamFoundationServer teamServer, int scanInterval )
        {
            m_teamServer = teamServer;
            m_scanInterval = scanInterval;
            m_versionControl = (VersionControlServer) m_teamServer.GetService( typeof( VersionControlServer ) );
        }

        private void BuildProject()
        {
            s_logger.InfoFormat( "****** Launching Build {0} ****", m_buildType );

            // Setup parameters
            BuildParameters buildParameters = new BuildParameters();
            buildParameters.TeamFoundationServer = m_teamServer.Uri.ToString();
            buildParameters.TeamProject = m_project.Name;
            buildParameters.BuildType = m_buildType;
            buildParameters.BuildDirectory = m_projectData.PropertyGroup.BuildDirectoryPath;
            buildParameters.BuildMachine = m_projectData.ProjectExtensions.BuildMachine;
            buildParameters.DropLocation = m_projectData.PropertyGroup.DropLocation;

            // Start build
            string buildUri;
            while( true )
            {
                try
                {
                    buildUri = Controller.StartBuild(buildParameters);
                    s_logger.InfoFormat("Build Started {0}", buildUri);
                    break;
                }
                catch (Exception ex)
                {
                    // If another build is in process then wait for it to finish.
                    if( ex.Message.Contains( "TF42045" ) )
                    {
                        Thread.Sleep( 5000 );
                    }
                    else
                    {
                        throw;
                    }
                }
            }

            // Wait until build finishes.
            BuildStore buildStore = (BuildStore) m_teamServer.GetService( typeof( BuildStore ) );
            while( true )
            {
                Thread.Sleep( 5000 );

                BuildData build = buildStore.GetBuildDetails( buildUri );
                if( build.FinishTime != DateTime.MinValue )
                {
                    break;
                }
            }
        }

        private IList<BuildData> GetBuildsList()
        {
            SortedList<DateTime, BuildData> retList = new SortedList<DateTime, BuildData>();
            foreach ( BuildData build in Store.GetListOfBuilds( m_project.Name, m_buildType ) )
            {
                retList.Add( build.StartTime, build );
            }

            return retList.Values;
        }

        private List<string> GetBuildTypes()
        {
            List<string> retList = new List<string>();

            string buildPath = VersionControlPath.Combine( m_project.ServerItem, "TeamBuildTypes/*" );
            foreach( Item buildItem in m_versionControl.GetItems( buildPath ).Items )
            {
                if( buildItem.ItemType == ItemType.Folder )
                {
                    string buildType = VersionControlPath.GetFileName( buildItem.ServerItem );
                    retList.Add( buildType );
                }
            }

            return retList;
        }

        private DateTime GetLastBuildDate()
        {
            IList<BuildData> buildList = GetBuildsList();
            if( buildList.Count > 0 )
            {
                return buildList[ buildList.Count - 1 ].FinishTime;
            }
            else
            {
                return DateTime.MinValue;
            }
        }

        private DateTime GetLatestChangesetDate( string serverItem )
        {
            IEnumerator enumChangeSets =
                m_versionControl.QueryHistory( serverItem, VersionSpec.Latest, 0, RecursionType.Full, null, null,
                                               null, 1, true, false ).GetEnumerator();
            while( enumChangeSets.MoveNext() )
            {
                Changeset lastChangeset = (Changeset) enumChangeSets.Current;

                if( lastChangeset.CreationDate > m_scanStartTime )
                {
                    continue;
                }

                return lastChangeset.CreationDate;
            }
            return m_scanStartTime;
        }

        private static XmlNode GetNode(XmlDocument xDoc, string tagName )
        {
            foreach( XmlNode node in xDoc.GetElementsByTagName( tagName ) )
            {
                return node;
            }
            return null;
        }

        private void ProcessBuildAtCheckInTag( XmlDocument xmlDoc )
        {
            XmlNode node = GetNode( xmlDoc, "BuildAtCheckIn" );
            if( node == null )
            {
                return;
            }
            s_logger.Info( "Found <BuildAtCheckIn> tag" );

            // Get latest change date among affected folders of build.
            bool doBuild = false;
            foreach( XmlNode mapNode in xmlDoc.GetElementsByTagName( "Map" ) )
            {
                XmlAttribute includePath = mapNode.Attributes[ "Include" ];
                s_logger.InfoFormat( "Found <Map> tag with value of {0}", includePath.InnerText );

                if( GetLatestChangesetDate( includePath.InnerText ) > GetLastBuildDate() )
                {
                    doBuild = true;
                    break;
                }
            }

            if( doBuild )
            {
                BuildProject();
            }
            else
            {
                s_logger.Info( "Build file does not need to be built" );
            }
        }

        private void ProcessBuildDailyAtTag( XmlDocument xmlDoc )
        {
            foreach ( XmlNode node in xmlDoc.GetElementsByTagName( "BuildDailyAt" ) )
            {
                s_logger.InfoFormat( "Found <BuildDailyAt> tag with value of {0}", node.InnerText );

                // Get start time for build
                DateTime startTime;
                if( !DateTime.TryParse( node.InnerText, out startTime ) )
                {
                    s_logger.WarnFormat( "Unable to parse date {0}", node.InnerText );
                    continue;
                }
                startTime = DateTime.Today + startTime.TimeOfDay;

                // If this build hasn't happened today and it's past due then build it.
                if( GetLastBuildDate() < DateTime.Today && startTime < m_scanStartTime )
                {
                    BuildProject();
                }
                else
                {
                    s_logger.Info( "Build file does not need to be built" );
                }
            }
        }

        private void ProcessBuildsToKeepTag( XmlDocument xmlDoc )
        {
            // Get builds to keep tag
            XmlNode node = GetNode( xmlDoc, "BuildsToKeep" );
            if( node == null )
            {
                return;
            }
            s_logger.InfoFormat( "Found <BuildsToKeep> tag with value of {0}", node.InnerText );

            // Parse into number.
            int buildsToKeep;
            if( !int.TryParse( node.InnerText, out buildsToKeep ) || buildsToKeep < 1 )
            {
                return;
            }

            // Get builds.
            IList<BuildData> buildList = GetBuildsList();

            s_logger.InfoFormat( "Current build count {0}, number to keep {1}", buildList.Count, buildsToKeep );
            if ( buildList.Count > buildsToKeep )
            {
                // Delete old builds.
                for( int x = 0 ; x < buildList.Count - buildsToKeep ; x++ )
                {
                    string errorMessage;
                    BuildData build = buildList[ x ];
                    s_logger.InfoFormat( "Deleting build {0}", build.BuildNumber );
                    if ( !Controller.DeleteBuild( build.BuildUri, out errorMessage ) )
                    {
                        s_logger.InfoFormat( "Error deleting build {0}: {1}", build.BuildNumber, errorMessage );
                        continue;
                    }

                    DirectoryInfo buildDir = new DirectoryInfo( build.DropLocation );
                    buildDir.Delete( true );
                }
            }
            else
            {
                s_logger.InfoFormat( "No builds to delete" );
            }
        }

        private static Project ReadBuildFile( string configFile )
        {
            using ( FileStream file = new FileStream( configFile, FileMode.Open, FileAccess.Read ) )
            {
                XmlSerializer serialize = new XmlSerializer( typeof( Project ) );
                return (Project) serialize.Deserialize( file );
            }
        }

        private void ScanBuildFile()
        {
            string projFilename = Path.GetTempFileName();

            try
            {
                s_logger.InfoFormat( "Processing build file {0}", m_buildFileItem.ServerItem );

                m_buildFileItem.DownloadFile( projFilename );
                XmlDocument xmlDoc = new XmlDocument();
                xmlDoc.Load( projFilename );
                m_projectData = ReadBuildFile( projFilename );

                ProcessBuildDailyAtTag( xmlDoc );
                ProcessBuildAtCheckInTag( xmlDoc );
                ProcessBuildsToKeepTag( xmlDoc );
            }
            catch( Exception ex )
            {
                s_logger.ErrorFormat( ex, "Error with build {0}.", m_buildType );
            }
            finally
            {
                File.Delete( projFilename );
            }
        }

        private void ScanBuildFiles()
        {
            foreach( string buildType in GetBuildTypes() )
            {
                string buildFilePath = VersionControlPath.Combine( VersionControlPath.Combine(  VersionControlPath.Combine( m_project.ServerItem, "TeamBuildTypes" ), buildType ), "TFSBuild.proj" );

                try
                {
                    m_buildType = buildType;
                    m_buildFileItem = m_versionControl.GetItem( buildFilePath );
                    ScanBuildFile();
                }
                catch ( Exception e )
                {
                    s_logger.ErrorFormat( e, "Exception Scanning Project {0} Build File {1}", m_project.Name, buildFilePath );
                }
            }
        }
    }
}