using System;
using System.Collections.Generic;
using System.IO;
using System.Xml;

using Microsoft.Build.Framework;
using Microsoft.Build.Utilities;

namespace Pegasus.MSBuild.Tasks.WixTask
{
	/// <summary>
	/// 
	/// </summary>
	public class Wix : Task
	{
		// Local Instance Values
		private string m_wxsFilename = null;
		private string m_outputFilename = null;
		private Version m_version = new Version( 1, 0 );
		private ITaskItem[] m_componentFiles = null;
		private ITaskItem[] m_wixLibraries = null;
		private string m_localizationFilename = null;
		private string m_wixDirectory = null;

		/// <summary>
		/// Initializes a new instance of the <see cref="T:Wix"/> class.
		/// </summary>
		public Wix()
		{
		}

		/// <summary>
		/// Gets or sets the filename.
		/// </summary>
		/// <value>The filename.</value>
		[Required]
		public string Filename
		{
			get
			{
				return m_wxsFilename;
			}
		
			set
			{
				m_wxsFilename = value;
			}
		}

		/// <summary>
		/// Gets or sets the output filename.
		/// </summary>
		/// <value>The output filename.</value>
		public string OutputFilename
		{
			get
			{
				return m_outputFilename;
			}

			set
			{
				m_outputFilename = value;
			}
		}

		/// <summary>
		/// Gets or sets the localization filename.
		/// </summary>
		/// <value>The localization filename.</value>
		public string LocalizationFilename
		{
			get
			{
				return m_localizationFilename;
			}

			set
			{
				m_localizationFilename = value;
			}
		}

		/// <summary>
		/// Gets or sets the version of the output file.
		/// </summary>
		/// <value>The version.</value>
		public string Version
		{
			get
			{
				return m_version.ToString();
			}

			set
			{
				m_version = new Version( value );
			}
		}

		/// <summary>
		/// Gets or sets the component files.
		/// </summary>
		/// <value>The component files.</value>
		public ITaskItem[] ComponentFiles
		{
			get
			{
				return m_componentFiles;
			}

			set
			{
				m_componentFiles = value;
			}
		}

		/// <summary>
		/// Gets or sets the wix libraries.
		/// </summary>
		/// <value>The wix libraries.</value>
		public ITaskItem[] Libraries
		{
			get
			{
				return m_wixLibraries;
			}

			set
			{
				m_wixLibraries = value;
			}
		}

		/// <summary>
		/// Gets or sets the wix directory.
		/// </summary>
		/// <value>The wix directory.</value>
		public string WixDirectory
		{
			get
			{
				return m_wixDirectory;
			}

			set
			{
				m_wixDirectory = value;
			}
		}

		/// <summary>
		/// When overridden in a derived class, executes the task.
		/// </summary>
		/// <returns>
		/// true if the task successfully executed; otherwise, false.
		/// </returns>
		public override bool Execute()
		{
			m_wixDirectory = Path.GetFullPath( m_wixDirectory );
			m_wxsFilename = Path.GetFullPath( m_wxsFilename );

			// Make sure that all the data is correct
			if( !WixCommon.ValidateWixData( Log, ref m_wixDirectory, m_wxsFilename ) )
			{
				return false;
			}

			Log.LogMessage( "Wix: Using Wix runtime files located at {0}", m_wixDirectory );

			// Build the output filename if needed
			if( string.IsNullOrEmpty( m_outputFilename ) )
			{
				m_outputFilename = Path.GetFileNameWithoutExtension( m_wxsFilename );
				m_outputFilename = Path.Combine( Path.GetDirectoryName( m_wxsFilename ), m_outputFilename + ".msi" );
			}

			m_outputFilename = Path.GetFullPath( m_outputFilename );

			// Process the wsx input file and make a copy of the modified file.
			string tempInputFilename = ProcessInputFile();
			if( tempInputFilename == null )
			{
				return false;
			}

			return BuildWixInstall( tempInputFilename );
		}

		/// <summary>
		/// Processes the input file.
		/// </summary>
		/// <returns></returns>
		private string ProcessInputFile()
		{
			try
			{
				Log.LogMessage( "Wix: Processing input file '{0}'", m_wxsFilename );

				XmlDocument xmlDocument = new XmlDocument();
				xmlDocument.Load( m_wxsFilename );

				// Get the xml namespace for the document
				string nameSpace = xmlDocument.DocumentElement.NamespaceURI;

				ProcessInputSetVersion( xmlDocument, nameSpace );
				ProcessInputAddFileTags( xmlDocument, nameSpace );

				// Create a temp file and same the XML document.
				string directory = Path.GetDirectoryName( m_wxsFilename );
				string filename = Path.GetFileName( m_wxsFilename );
				string tempFilename = Path.Combine( directory, "Temp_" + filename );

				xmlDocument.Save( tempFilename );
				
				return tempFilename;
			}
			catch( Exception e )
			{
				Log.LogError( "Wix: {0}", e.Message );
			}

			return null;
		}

		/// <summary>
		/// Processes the input set version.
		/// </summary>
		/// <param name="xmlDocument">The XML document.</param>
		/// <param name="nameSpace">The name space.</param>
		public void ProcessInputSetVersion( XmlDocument xmlDocument, string nameSpace )
		{
			// Process the version attribute on all the Product tags
			XmlNodeList xmlProductList = xmlDocument.GetElementsByTagName( "Product", nameSpace );
			foreach( XmlNode xmlProduct in xmlProductList )
			{
				XmlAttribute versionAttrib = xmlProduct.Attributes[ "Version" ];
				if( versionAttrib != null )
				{
					versionAttrib.Value = m_version.ToString();
					Log.LogMessage( "Wix: Setting Version {0}", m_version.ToString() );
				}
			}
		}

		/// <summary>
		/// Processes the input add file tags.
		/// </summary>
		/// <param name="xmlDocument">The XML document.</param>
		/// <param name="nameSpace">The name space.</param>
		public void ProcessInputAddFileTags( XmlDocument xmlDocument, string nameSpace )
		{
			// Used to contain a list of unique 8.3 filenames
			List<string> uniqueIds = new List<string>();
			List<string> uniqueFilenames = new List<string>();

			XmlNodeList xmlComponentList = xmlDocument.GetElementsByTagName( "Component", nameSpace );
			if( xmlComponentList != null )
			{
				foreach( XmlNode xmlComponent in xmlComponentList )
				{
					string componentName = xmlComponent.Attributes[ "Id" ].Value;
					
					Dictionary< string, ITaskItem > componentFiles = GetComponentFiles( componentName );

					if( componentFiles.Count > 0 )
					{
						// Process any existing file nodes
						foreach( XmlNode node in xmlComponent.ChildNodes )
						{
							if( node.Name == "File" )
							{
								// Get the filename for the
								string id = node.Attributes[ "Id" ].Value;
								string filename = Path.GetFileName( node.Attributes[ "LongName" ].Value ).ToLower();

								if( componentFiles.ContainsKey( filename ) )
								{
									ITaskItem item = componentFiles[ filename ];
									node.Attributes[ "src" ].Value = item.ItemSpec;
									
									// We have used this file so remove it from the table.
									componentFiles.Remove( filename );

									Log.LogMessage( "Wix: Updated File tag {0} {1} src = '{2}'", componentName, id, item.ItemSpec );
								}
							}
						}

						// Create a tag for each file that looks like this
						// <File Id="TheFile" Name="TheFile.bld" 
						//		LongName="TheFile.build" src="c:\mydir\TheFile.build" />

						foreach( ITaskItem item in componentFiles.Values )
						{
							string longFilename = Path.GetFileName( item.ItemSpec );
							string shortFilename = GetEightDotThreeFilename( longFilename, uniqueFilenames );
							string idFilename = GetIdFilename( longFilename, uniqueIds );

							XmlElement fileNode = xmlDocument.CreateElement( "File" );

							XmlAttribute idAttrib = xmlDocument.CreateAttribute( "Id" );
							idAttrib.Value = idFilename;
							fileNode.Attributes.Append( idAttrib );

							XmlAttribute nameAttrib = xmlDocument.CreateAttribute( "Name" );
							nameAttrib.Value = shortFilename;
							fileNode.Attributes.Append( nameAttrib );

							XmlAttribute longNameAttrib = xmlDocument.CreateAttribute( "LongName" );
							longNameAttrib.Value = longFilename;
							fileNode.Attributes.Append( longNameAttrib );

							XmlAttribute srcAttrib = xmlDocument.CreateAttribute( "src" );
							srcAttrib.Value = item.ItemSpec;
							fileNode.Attributes.Append( srcAttrib );

							xmlComponent.AppendChild( fileNode );

							Log.LogMessage( "Wix: Added File tag {0} {1} src = '{2}'", componentName, idFilename, item.ItemSpec );
						}
					}
				}
			}
		}

		/// <summary>
		/// Gets the component files.
		/// </summary>
		/// <param name="name">The name.</param>
		/// <returns></returns>
		private Dictionary<string, ITaskItem> GetComponentFiles( string name )
		{
			Dictionary<string, ITaskItem> files = new Dictionary<string, ITaskItem>();

			if( m_componentFiles != null )
			{
				foreach( ITaskItem item in m_componentFiles )
				{
					if( item.GetMetadata( "Component" ) == name )
					{
						files.Add( Path.GetFileName( item.ItemSpec ).ToLower(), item );
					}
				}
			}

			return files;
		}

		/// <summary>
		/// Gets the id filename.
		/// </summary>
		/// <param name="filename">The filename.</param>
		/// <param name="usedIds">The used ids.</param>
		/// <returns></returns>
		private string GetIdFilename( string filename, List<string> usedIds )
		{
			bool done = false;
			string orginalId = filename.Replace( ".", "_" ).Replace( "-", "_" );
			int count = 0;

			string id = orginalId;
			while( !done )
			{
				if( usedIds.Contains( id ) )
				{
					count++;
					id = orginalId + count.ToString();
				}
				else
				{
					usedIds.Add( id );
					done = true;
				}
			}

			return id;
		}

		/// <summary>
		/// Gets the eight dot three filename.
		/// </summary>
		/// <param name="longFilename">The long filename.</param>
		/// <param name="usedFilenames">The used filenames.</param>
		/// <returns></returns>
		public string GetEightDotThreeFilename( string longFilename, List<string> usedFilenames )
		{
			bool done = false;
			string uniqueAppend = string.Empty;
			string shortFilename = string.Empty;
			int count = 0;

			while( !done )
			{
				shortFilename = GetEightDotThreeFilename( longFilename, uniqueAppend );
				if( usedFilenames.Contains( shortFilename ) )
				{
					count++;
					uniqueAppend = count.ToString();
				}
				else
				{
					usedFilenames.Add( shortFilename );
					done = true;
				}
			}

			return shortFilename;
		}

		/// <summary>
		/// Gets the eight dot three filename.
		/// </summary>
		/// <param name="longFilename">The long filename.</param>
		/// <param name="uniqueAppend">The unique append.</param>
		/// <returns></returns>
		private string GetEightDotThreeFilename( string longFilename, string uniqueAppend )
		{
			string filename = Path.GetFileNameWithoutExtension( longFilename );
			string ext = Path.GetExtension( longFilename );

			filename = filename.Replace( ".", "" );

			int len = Math.Min( 8 - uniqueAppend.Length, filename.Length );
			len = Math.Max( 0, len );
			filename = filename.Substring( 0, len );

			len = Math.Min( 3, ext.Length - 1 );
			len = Math.Max( 0, len );
			ext = ext.Substring( 1, len );

			return string.Format( "{0}{1}.{2}", filename, uniqueAppend, ext );
		}

		/// <summary>
		/// Builds the wix install.
		/// </summary>
		/// <param name="wxsFilename">The WXS filename.</param>
		/// <returns></returns>
		private bool BuildWixInstall( string wxsFilename )
		{
			// Build up the candle task
			WixCandle candleTask = new WixCandle();
			candleTask.HostObject = HostObject;
			candleTask.BuildEngine = BuildEngine;

			candleTask.Filename = wxsFilename;
			candleTask.WixDirectory = m_wixDirectory;

			// Execute the candle task
			if( !candleTask.Execute() )
			{
				return false;
			}

			string objFilename = candleTask.OutputFilename;

			// Build up the light task
			WixLight lightTask = new WixLight();
			lightTask.HostObject = HostObject;
			lightTask.BuildEngine = BuildEngine;
			
			lightTask.Filename = objFilename;
			lightTask.OutputFilename = m_outputFilename;
			lightTask.Libraries = m_wixLibraries;
			lightTask.LocalizationFilename = m_localizationFilename;
			lightTask.WixDirectory = m_wixDirectory;

			return lightTask.Execute();
		}
	}
}
