using System;
using System.IO;

using Microsoft.Build.Utilities;

namespace Pegasus.MSBuild.Tasks.WixTask
{
	/// <summary>
	/// 
	/// </summary>
	internal static class WixCommon
	{
		/// <summary>
		/// Validates the input data.
		/// </summary>
		/// <returns></returns>
		public static bool ValidateWixData( TaskLoggingHelper log, ref string wixDirectory, string wxsFilename )
		{
			// Locate and validate the Wix environment
			if( string.IsNullOrEmpty( wixDirectory ) )
			{
				wixDirectory = WixCommon.FindWixDirectory();
				if( wixDirectory == null )
				{
					log.LogError( "WixCommon: Unable to find the Wix runtime environment." );
					return false;
				}
			}
			else if( !WixCommon.IsWixDirectory( wixDirectory ) )
			{
				log.LogError( "WixCommon: '{0}' is not a valid directory or does not contain the Wix runtime files.", wixDirectory );
				return false;
			}

			// Validate the wix input file
			if( !File.Exists( wxsFilename ) )
			{
				log.LogError( "WixCommon: '{0}' does not exist or can not be accessed.", wxsFilename );
				return false;
			}

			return true;
		}


		/// <summary>
		/// Finds the wix directory.
		/// </summary>
		/// <returns></returns>
		private static string FindWixDirectory()
		{
			// See if the WIXDIR environment value is set.
			string wixDir = Environment.ExpandEnvironmentVariables( "%wixdir%" );
			if( WixCommon.IsWixDirectory( wixDir ) )
			{
				return wixDir;
			}

			// Is wix under the program file directory.
			wixDir = Environment.GetFolderPath( Environment.SpecialFolder.ProgramFiles );
			wixDir = Path.Combine( wixDir, "wix" );
			if( WixCommon.IsWixDirectory( wixDir ) )
			{
				return wixDir;
			}

			return null;
		}

		/// <summary>
		/// Determines whether this given directory holds the WIX runtime files.
		/// </summary>
		/// <param name="wixDir">The wix dir.</param>
		/// <returns>
		/// 	<c>true</c> if [is wix directory] [the specified wix dir]; otherwise, <c>false</c>.
		/// </returns>
		private static bool IsWixDirectory( string wixDir )
		{
			// Is there are path to test
			if( string.IsNullOrEmpty( wixDir ) )
			{
				return false;
			}

			// Does the directory exist
			if( !Directory.Exists( wixDir ) )
			{
				return false;
			}

			// Does it have the candle.exe file in it
			string candleExe = Path.Combine( wixDir, "candle.exe" );
			if( !File.Exists( candleExe ) )
			{
				return false;
			}

			// Does it have the light.exe file in it
			string lightExe = Path.Combine( wixDir, "light.exe" );
			if( !File.Exists( candleExe ) )
			{
				return false;
			}

			// Does the ca directory exist
			string caDir = Path.Combine( wixDir, "ca" );
			if( !Directory.Exists( caDir ) )
			{
				return false;
			}

			// Does the include directory exist
			string includeDir = Path.Combine( wixDir, "include" );
			if( !Directory.Exists( includeDir ) )
			{
				return false;
			}

			return true;
		}
	}
}
