using System;
using System.IO;

using Microsoft.Build.Framework;
using Microsoft.Build.Utilities;

namespace Pegasus.MSBuild.Tasks.WixTask
{
	/// <summary>
	/// 
	/// </summary>
	public class WixLight : ToolTask
	{
		// Local Instance Values
		private string m_wxsFilename = null;
		private string m_outputFilename = null;
		private ITaskItem[] m_wixLibraries = null;
		private string m_localizationFilename = null;
		
		private string m_wixDirectory = null;

		/// <summary>
		/// Initializes a new instance of the <see cref="T:WixLight"/> class.
		/// </summary>
		public WixLight()
		{
		}

		/// <summary>
		/// Gets or sets the filename.
		/// </summary>
		/// <value>The filename.</value>
		[Required]
		public string Filename
		{
			get
			{
				return m_wxsFilename;
			}

			set
			{
				m_wxsFilename = value;
			}
		}

		/// <summary>
		/// Gets or sets the filename.
		/// </summary>
		/// <value>The filename.</value>
		public string OutputFilename
		{
			get
			{
				return m_outputFilename;
			}

			set
			{
				m_outputFilename = value;
			}
		}

		/// <summary>
		/// Gets or sets the libraries.
		/// </summary>
		/// <value>The libraries.</value>
		public ITaskItem[] Libraries
		{
			get
			{
				return m_wixLibraries;
			}
		
			set
			{
				m_wixLibraries = value;
			}
		}

		/// <summary>
		/// Gets or sets the localization filename.
		/// </summary>
		/// <value>The localization filename.</value>
		public string LocalizationFilename
		{
			get
			{
				return m_localizationFilename;
			}

			set
			{
				m_localizationFilename = value;
			}
		}

		/// <summary>
		/// Gets or sets the wix directory.
		/// </summary>
		/// <value>The wix directory.</value>
		public string WixDirectory
		{
			get
			{
				return m_wixDirectory;
			}

			set
			{
				m_wixDirectory = value;
			}
		}

		/// <summary>
		/// Gets the name of the executable file to run.
		/// </summary>
		/// <value></value>
		/// <returns>The name of the executable file to run.</returns>
		protected override string ToolName
		{
			get
			{
				return "light.exe";
			}
		}

		/// <summary>
		/// Runs the exectuable file with the specified task parameters.
		/// </summary>
		/// <returns>
		/// true if the task runs successfully; otherwise, false.
		/// </returns>
		public override bool Execute()
		{
			m_wixDirectory = Path.GetFullPath( m_wixDirectory );
			m_wxsFilename = Path.GetFullPath( m_wxsFilename );

			// Make sure that all the data is correct
			if( !WixCommon.ValidateWixData( Log, ref m_wixDirectory, m_wxsFilename ) )
			{
				return false;
			}

			Log.LogMessage( "WixLight: Using Wix runtime files located at {0}", m_wixDirectory );

			// Build the output filename if needed
			if( string.IsNullOrEmpty( m_outputFilename ) )
			{
				m_outputFilename = Path.GetFileNameWithoutExtension( m_wxsFilename );
				m_outputFilename = Path.Combine( Path.GetDirectoryName( m_wxsFilename ), m_outputFilename + ".msi" );
			}

			m_outputFilename = Path.GetFullPath( m_outputFilename );

			Log.LogMessage( "WixLight: Linking Wix file {0}", m_wxsFilename );
			return base.Execute();
		}

		/// <summary>
		/// Returns the fully qualified path to the executable file.
		/// </summary>
		/// <returns>
		/// The fully qualified path to the executable file.
		/// </returns>
		protected override string GenerateFullPathToTool()
		{
			return Path.Combine( m_wixDirectory, ToolName );
		}

		/// <summary>
		/// Returns a string value containing the command line arguments to pass directly to the executable file.
		/// </summary>
		/// <returns>
		/// A string value containing the command line arguments to pass directly to the executable file.
		/// </returns>
		protected override string GenerateCommandLineCommands()
		{
			CommandLineBuilder builder = new CommandLineBuilder();

			// We don't need the tool's logo information shown
			builder.AppendSwitch( "-nologo" );

			// Set the input file
			builder.AppendFileNameIfNotNull( m_wxsFilename );

			// Set the output file
			if( !string.IsNullOrEmpty( m_outputFilename ) )
			{
				builder.AppendSwitch( "-out" );
				builder.AppendFileNameIfNotNull( m_outputFilename );
			}

			// Add Wix ca libraries to the build 
			if( m_wixLibraries != null )
			{
				string caDirectory = Path.Combine( m_wixDirectory, "ca" );

				foreach( ITaskItem item in m_wixLibraries )
				{
					string libFilename = Path.Combine( caDirectory, item.ItemSpec );
					libFilename = Path.GetFullPath( libFilename );
					builder.AppendFileNameIfNotNull( libFilename );
				}
			}

			// Set the language file
			if( !string.IsNullOrEmpty( m_localizationFilename ) )
			{
				builder.AppendSwitch( "-loc" );
				builder.AppendFileNameIfNotNull( m_localizationFilename );
			}

			return builder.ToString();
		}
	}
}
