using System;
using System.IO;

using Microsoft.Build.Framework;
using Microsoft.Build.Utilities;
using Microsoft.TeamFoundation.Client;
using Microsoft.TeamFoundation.Build.Client;
using Microsoft.TeamFoundation.VersionControl.Common;
using Microsoft.TeamFoundation.VersionControl.Client;
using System.Collections.Generic;

namespace Pegasus.MSBuild.TeamFoundation.Tasks.VersionControlTask
{
	/// <summary>
	/// This task will delete old finished build files.
	/// </summary>
	public class DeleteOldBuildOutput : VersionTasksBase
	{
		// Local Instance Values
		private int m_buildsToKeep;
        private string m_projectName;
        private string m_buildDefName;
        private bool m_deleteBuildInfo = false;
        private IBuildDefinition m_buildDef;
        private IBuildServer m_buildServer;
        private VersionControlServer m_versionServer;

        /// <summary>
        /// Gets for sets the build definition name.
        /// </summary>
        [Required]
        public string BuildDefinitionName
        {
            get
            {
                return m_buildDefName;
            }
            set
            {
                m_buildDefName = value;
            }
        }

		/// <summary>
		/// Gets or sets the number of builds to keep.
		/// </summary>
		[Required]
        public int BuildsToKeep
		{
			get
			{
                return m_buildsToKeep;
			}

			set
			{
                m_buildsToKeep = value;
			}
		}

        /// <summary>
        /// Gets or sets whether to delete old builds' information from TFS as well.  The default is false.
        /// </summary>
        public bool DeleteBuildInfo
        {
            get
            {
                return m_deleteBuildInfo;
            }
            set
            {
                m_deleteBuildInfo = value;
            }
        }

        /// <summary>
        /// Gets or sets the name of the project.
        /// </summary>
        [Required]
        public string TeamProject
        {
            get
            {
                return m_projectName;
            }

            set
            {
                m_projectName = value;
            }
        }

        private IBuildDefinition BuildDefinition
        {
            get
            {
                if( m_buildDef == null )
                {
                    m_buildDef = BuildServer.GetBuildDefinition( TeamProject, BuildDefinitionName );
                }
                return m_buildDef;
            }
        }

        private IBuildServer BuildServer
        {
            get
            {
                if( m_buildServer == null )
                {
                    m_buildServer = (IBuildServer) m_versionServer.TeamFoundationServer.GetService( typeof( IBuildServer ) );
                }
                return m_buildServer;
            }
        }

        protected override bool ProcessFiles( VersionControlServer versionServer )
        {
            m_versionServer = versionServer;

            // Get build's drop location.
            DirectoryInfo dropLocation = new DirectoryInfo( BuildDefinition.DefaultDropLocation );
            if( !dropLocation.Exists )
            {
                Log.LogWarning( "Builds default drop location doesn't exist so the tast will be skipped." );
                return false;
            }

            // Get sub-folders that are build outputs.
            SortedList<DateTime, IBuildDetail> buildsWithOutput = new SortedList<DateTime, IBuildDetail>();
            foreach( DirectoryInfo subDir in dropLocation.GetDirectories() )
            {
                try
                {
                    IBuildDetail build = BuildServer.GetBuild( BuildDefinition.CreateSpec(),
                        subDir.Name, new string[] { }, QueryOptions.None );

                    buildsWithOutput.Add( build.StartTime, build );
                }
                catch( Exception ex )
                {
                    if( ex.Message.Contains( "No build was found with build number" ) )
                    {
                        // folder was encountered that isn't a build.  Ignore it.
                        continue;
                    }
                }
            }

            // Check if that are too many builds.
            if( buildsWithOutput.Count <= BuildsToKeep )
            {
                Log.LogMessage( "{0} builds found and {1} builds allowed.  No action necessary.",
                    buildsWithOutput.Count, BuildsToKeep );
                return true;
            }

            // Build list of builds that need deleting.
            List<IBuildDetail> buildsToDelete = new List<IBuildDetail>();
            while( buildsWithOutput.Count > BuildsToKeep )
            {
                buildsToDelete.Add( buildsWithOutput.Values[ 0 ] );
                buildsWithOutput.RemoveAt( 0 );
            }

            // Delete old build outputs.
            foreach( IBuildDetail oldBuild in buildsToDelete )
            {
                Log.LogMessage( "Deleting build {0} output.", oldBuild.BuildNumber );

                try
                {
                    Directory.Delete( oldBuild.DropLocation, true );
                }
                catch( Exception ex )
                {
                    Log.LogWarningFromException( ex, true );
                }
            }

            // Delete build log if specified.
            if( DeleteBuildInfo )
            {
                foreach( IBuildDeletionResult result in BuildServer.DeleteBuilds( buildsToDelete.ToArray() ) )
                {
                    logDeleteResult( result );
                }
            }

            return true;
        }

        private void logDeleteResult( IBuildDeletionResult result )
        {
            if( result.Successful )
            {
                Log.LogMessage( "Deleting build results." );
            }
            else
            {
                Log.LogError( "Unable to delete build results" );
                if( result.DropLocationFailure != null )
                {
                    Log.LogError( "Drop Location Failure {0}: {1}", 
                        result.DropLocationFailure.Code, result.DropLocationFailure.Message );
                }

                if( result.LabelFailure != null )
                {
                    Log.LogError( "Label Failure {0}: {1}", 
                        result.LabelFailure.Code, result.LabelFailure.Message );
                }

                if( result.TestResultFailure != null )
                {
                    Log.LogError( "Test Result Failure {0}: {1}", 
                        result.TestResultFailure.Code, result.TestResultFailure.Message );
                }
            }
        }

        private IBuildDetail getBuild( string buildName )
        {
            return BuildServer.GetBuild( BuildDefinition.CreateSpec(), buildName, 
                new string[] { }, QueryOptions.None );
        }
    }
}
