using System;
using System.IO;

using Microsoft.Build.Framework;
using Microsoft.Build.Utilities;
using Microsoft.TeamFoundation.Client;
using Microsoft.TeamFoundation.VersionControl.Common;
using Microsoft.TeamFoundation.VersionControl.Client;

namespace Pegasus.MSBuild.TeamFoundation.Tasks.VersionControlTask
{
	/// <summary>
	/// This task will get the lastest version of a file/directory and copy it to the given location.
	/// </summary>
	public class GetFiles : VersionTasksBase
	{
		// Local Instance Values
		private string m_serverPath = null;
		private string m_localPath = null;
		private bool m_recursive = false;

		/// <summary>
		/// Initializes a new instance of the <see cref="T:GetFile"/> class.
		/// </summary>
		public GetFiles()
		{
		}

		/// <summary>
		/// Gets or sets the server path.
		/// </summary>
		/// <value>The server path.</value>
		[Required]
		public string ServerPath
		{
			get
			{
				return m_serverPath;
			}

			set
			{
				m_serverPath = value;
			}
		}

		/// <summary>
		/// Gets or sets the local path.  Defaults to the Current Working Directory.
		/// </summary>
		/// <value>The local path.</value>
		public string LocalPath
		{
			get
			{
				return m_localPath;
			}

			set
			{
				m_localPath = value;
			}
		}

		/// <summary>
		/// Gets or sets a value indicating whether this <see cref="T:GetFiles"/> is recursive.
		/// </summary>
		/// <value><c>true</c> if recursive; otherwise, <c>false</c>.</value>
		public bool Recursive
		{
			get
			{
				return m_recursive;
			}

			set
			{
				m_recursive = value;
			}
		}

		/// <summary>
		/// Processes the files.
		/// </summary>
		/// <param name="versionServer">The version server.</param>
		/// <param name="outputDir">The output dir.</param>
		/// <returns></returns>
		protected override bool ProcessFiles( VersionControlServer versionServer )
		{
			int parentId = 0;

			// Default to the current working directory.
			string outputDir = m_localPath;
			if( string.IsNullOrEmpty( outputDir ) )
			{
				outputDir = Environment.CurrentDirectory;
			}

			outputDir = Path.GetFullPath( outputDir );

			ItemSet serverItems = versionServer.GetItems( m_serverPath, RecursionType.None );
			
			// If the item is a directory then get the sub item
			if( serverItems.Items.Length == 1 && serverItems.Items[ 0 ].ItemType == ItemType.Folder )
			{
				parentId = serverItems.Items[ 0 ].ItemId;
				serverItems = versionServer.GetItems( serverItems.Items[ 0 ].ServerItem, RecursionType.OneLevel );
			}

			DownloadFiles( versionServer, serverItems.Items, parentId, outputDir, m_recursive );
			
			return true;
		}

		/// <summary>
		/// Downloads the files.
		/// </summary>
		/// <param name="versionServer">The version server.</param>
		/// <param name="items">The items.</param>
		/// <param name="outputDir">The output dir.</param>
		/// <param name="recursive">if set to <c>true</c> [recursive].</param>
		private void DownloadFiles( VersionControlServer versionServer, Item[] items, int parentId, string outputDir, bool recursive )
		{
			foreach( Item item in items )
			{
				if( item.ItemType == ItemType.File )
				{
					// Build the output filename
					string outputFilename = Path.Combine( outputDir, VersionControlPath.GetFileName( item.ServerItem ) );

					// Download the file from the version control server
					Log.LogMessage( "GetFile: Downloading {0} to {1}", item.ServerItem, outputFilename );

					versionServer.DownloadFile( item.ServerItem, outputFilename );
				}
				else if( item.ItemType == ItemType.Folder && recursive && item.ItemId != parentId )
				{
					string subDir = Path.Combine( outputDir, VersionControlPath.GetFileName( item.ServerItem ) );
					if( !Directory.Exists( subDir ) )
					{
						Directory.CreateDirectory( subDir );
					}

					ItemSet serverItems = versionServer.GetItems( item.ServerItem, RecursionType.OneLevel );
					DownloadFiles( versionServer, serverItems.Items, item.ItemId, subDir, recursive );
				}
			}
		}
	}
}
