using System;
using Microsoft.Build.Framework;
using Microsoft.TeamFoundation.Build.Client;
using Microsoft.TeamFoundation.VersionControl.Client;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace Pegasus.MSBuild.TeamFoundation.Tasks.VersionControlTask
{
	/// <summary>
	/// This task will log all the change sets and work items of a build
	/// </summary>
	public class LogChangeSets : VersionTasksBase
	{
		// Local Instance Values
        private string m_buildUri;
        private string m_outputFile;
        private IBuildServer m_buildServer;
        private VersionControlServer m_versionServer;

        /// <summary>
        /// Gets or sets the build uri.
        /// </summary>
        [Required]
        public string BuildUri
        {
            get
            {
                return m_buildUri;
            }
            set
            {
                m_buildUri = value;
            }
        }

        /// <summary>
        /// Gets or sets the output file name for the logging.
        /// </summary>
        [Required]
        public string OutputFile
        {
            get
            {
                return m_outputFile;
            }
            set
            {
                m_outputFile = value;
            }
        }

        private IBuildServer BuildServer
        {
            get
            {
                if( m_buildServer == null )
                {
                    m_buildServer = (IBuildServer) m_versionServer.TeamFoundationServer.GetService( typeof( IBuildServer ) );
                }
                return m_buildServer;
            }
        }

        protected override bool ProcessFiles( VersionControlServer versionServer )
        {
            m_versionServer = versionServer;

            // Get build
            IBuildDetail build = GetBuild();

            // Extract change sets and work items.
            List<Dictionary<string, string>> changesets = new List<Dictionary<string, string>>();
            List<Dictionary<string, string>> workitems = new List<Dictionary<string, string>>();
            foreach( IBuildInformationNode info in build.Information.Nodes )
            {
                switch( info.Type )
                {
                    case "AssociatedChangeset":
                        changesets.Add( info.Fields );
                        break;

                    case "AssociatedWorkItem":
                        workitems.Add( info.Fields );
                        break;

                    default:
                        break;
                }
            }

            // Create log output.
            StringBuilder sb = new StringBuilder();

            if( changesets.Count > 0 )
            {
                sb.AppendLine( "Associated changesets:" );
                foreach( Dictionary<string, string> item in changesets )
                {
                    sb.AppendFormat( "ID: {0},\tChecked In By: {1},\tComments: {2}\n",
                        item[ "ChangesetId" ], item[ "CheckedInBy" ], item[ "Comment" ] );
                }
                sb.AppendLine();
            }

            if( workitems.Count > 0 )
            {
                sb.AppendLine( "Associated work items:" );
                foreach( Dictionary<string, string> item in workitems )
                {
                    sb.AppendFormat( "ID: {0},\tTitle: {1},\tState: {2},\tAssigned To: {3}\n",
                        item[ "WorkItemId" ], item[ "Title" ], item[ "Status" ], item[ "AssignedTo" ] );
                }
            }

            File.AppendAllText( OutputFile, sb.ToString() );
            return true;
        }

        private void logDeleteResult( IBuildDeletionResult result )
        {
            if( result.Successful )
            {
                Log.LogMessage( "Deleting build results." );
            }
            else
            {
                Log.LogError( "Unable to delete build results" );
                if( result.DropLocationFailure != null )
                {
                    Log.LogError( "Drop Location Failure {0}: {1}", 
                        result.DropLocationFailure.Code, result.DropLocationFailure.Message );
                }

                if( result.LabelFailure != null )
                {
                    Log.LogError( "Label Failure {0}: {1}", 
                        result.LabelFailure.Code, result.LabelFailure.Message );
                }

                if( result.TestResultFailure != null )
                {
                    Log.LogError( "Test Result Failure {0}: {1}", 
                        result.TestResultFailure.Code, result.TestResultFailure.Message );
                }
            }
        }

        private IBuildDetail GetBuild()
        {
            return BuildServer.GetAllBuildDetails( new Uri( BuildUri ) );
        }
    }
}
