using System;
using System.IO;

using Microsoft.Build.Framework;
using Microsoft.Build.Utilities;
using Microsoft.TeamFoundation.Client;
using Microsoft.TeamFoundation.VersionControl.Common;
using Microsoft.TeamFoundation.VersionControl.Client;

namespace Pegasus.MSBuild.TeamFoundation.Tasks.VersionControlTask
{
	/// <summary>
	/// This is the base class/common code for all the Version Control Tasks
	/// </summary>
	public abstract class VersionTasksBase : Task
	{
		// Local Instance Values
		private string m_serverUrl = null;

		/// <summary>
		/// Initializes a new instance of the <see cref="T:VersionTasksBase"/> class.
		/// </summary>
		public VersionTasksBase()
		{
		}

		/// <summary>
		/// Gets or sets the name of the team foundation server.
		/// </summary>
		/// <value>The server.</value>
		public string TeamFoundationServerUrl
		{
			get
			{
				return m_serverUrl;
			}

			set
			{
				m_serverUrl = value;
			}
		}

		/// <summary>
		/// When overridden in a derived class, executes the task.
		/// </summary>
		/// <returns>
		/// true if the task successfully executed; otherwise, false.
		/// </returns>
		public override bool Execute()
		{
			return ProcessFiles( TeamServerReference.GetTeamFoundationVersionControl( m_serverUrl ) );
		}

		/// <summary>
		/// Processes the files.
		/// </summary>
		/// <param name="versionServer">The version server.</param>
		/// <returns></returns>
		protected abstract bool ProcessFiles( VersionControlServer versionServer );

		/// <summary>
		/// Gets the workspace.
		/// </summary>
		/// <param name="versionServer">The version server.</param>
		/// <param name="filename">The filename.</param>
		/// <returns></returns>
		protected Workspace GetWorkspace( VersionControlServer versionServer, string filename )
		{
			WorkspaceInfo workspaceInfo = GetWorkspaceInfo( filename );
			return workspaceInfo.GetWorkspace( versionServer.TeamFoundationServer );
		}

		/// <summary>
		/// Gets the workspace info.
		/// </summary>
		/// <param name="filename">The filename.</param>
		/// <returns></returns>
		protected WorkspaceInfo GetWorkspaceInfo( string filename )
		{
			if( !Workstation.Current.IsMapped( filename ) )
			{
				throw new Exception( "No workspace mapping for filename." );
			}

			return Workstation.Current.GetLocalWorkspaceInfo( filename );
		}
	}
}
